#   toba -- build a Java program using Toba
#
#   Copyright 1996 Arizona Board of Regents; see COPYRIGHT file for details.
#
#
#   usage:  toba [options] file...
#
#   options:
#
#	-X		use directly executable version of Toba (default)
#	-J		use interpreted version of Toba
#
#	-K		run javac to produce .class files but go no further
#	-C		produce C files but don't compile or link them
#	-S		produce .s files but don't assemble or link them
#	-c		compile to .o files but do not link
#	-o filename	write executable to filename instead of a.out
#	-R libdir	embed specified library directory instead of default
#
#	-I path		set include path for C compiler
#			(note: Java API is always in path)
#	-g		pass -g (enable debugging) to cc
#	-O		pass -O (enable optimization) to cc
#	-A string	add string to args passed to cc
#
#	-i		retain intermediate files upon exit
#	-v 		("verbose") list commands executed
#
#	-d flags	set debugging flags for Toba
#
#
#   file argument interpretation:
#
#	file.java	Java file to compile, translate, and link
#	file.class	class file to translate and link
#	file.c		C file to compile and link
#	file.o		object file to link
#	file.a		library to link
#	file.so		library to link
#	-lxxx		library to link (must follow a "normal" file)
#	classname	Java class from $CLASSPATH to translate and link


# explicit or defaulted class path, before adding what we need
CPATH="${CLASSPATH:-.}"

# deduce current directory
WHENCE=`whence $0`
case $WHENCE in
    /*)	;;
    *)	WHENCE=`pwd`/$WHENCE;;
esac
PROG=`echo $WHENCE | sed 's=/[^/][^./]*/\.\./=/=g'`
TOP=${PROG%/*/*}
JDK=$TOP/jdk

# paths for interpreted and executed javac and toba
IJAVAC=$JDK/bin/javac
ITOBA=$TOP/bin/itoba
XJAVAC="LD_LIBRARY_PATH=$TOP/lib $TOP/bin/xjavac"
XTOBA="LD_LIBRARY_PATH=$TOP/lib $TOP/bin/xtoba"

# standard libraries for linking
RLIBDIR="$TOP/lib"

# default option values
TOBA=$XTOBA
JAVAC=$XJAVAC
NOTOBA=
NOCOMPILE=
NOLINK=
XFILE=a.out
DBGFLAGS=
INCLUDES="-I$TOP/include"
ECHO=:
CFLAGS="$CTOBA -w"
LDFLAGS="-L $TOP/lib"


#  execute command, possibly echoing to stdout
function trace {
    trap 'rm -f $TMP; exit 1' 1 2 15	# seems to be needed
    # sed protects dollar signs from expansion; tr protects sed from long lines
    CMD=`echo "$@" | tr ' ' '\012' | sed 's/\\$/\\\\$/g' | tr '\012' ' '`
    $ECHO + "$CMD"
    eval "$CMD" || exit 1
}

#  remove files; disabled by -i option
function remove {
    trap 'rm -f $TMP; exit 1' 1 2 15	# seems to be needed
    if [ $# -gt 0 ]; then
	$RMTRACE rm -f "$@"
    fi
}
RMTRACE=trace



#  process command options

USAGE="usage: $0 [-cCSKgOivJX] [-o file] [-IR dir] [-A arg] file..."

while getopts cCSKgOivJXo:R:I:A:d: c; do
    case $c in
	c)  NOLINK=exit;;
	C)  NOCOMPILE=exit;;
	S)  NOLINK=exit;  CFLAGS="$CFLAGS -S";;
	K)  NOTOBA=exit;;
	g)  CFLAGS="$CFLAGS -g";;
	O)  CFLAGS="$CFLAGS -O";;
	i)  RMTRACE=:;;
	v)  ECHO=echo;;
	J)  TOBA=$ITOBA; JAVAC=$IJAVAC;;
	X)  TOBA=$XTOBA; JAVAC=$XJAVAC;;
	o)  XFILE="$OPTARG";;
	I)  INCLUDES="$INCLUDES -I$OPTARG";;
	A)  CFLAGS="$CFLAGS $OPTARG";;
	d)  DBGFLAGS="-d$OPTARG";;
	R)  RLIBDIR="$OPTARG";;
	\?) echo $USAGE; exit 1;;
    esac
done


#  process file names

shift `expr $OPTIND - 1`

if [ $# = 0 ]; then
    echo $USAGE
    exit 1
fi

JFILES=
KFILES=
CFILES=
LFILES="-ltoba $LIBS"

for f; do
    case $f in
	*.java)		JFILES="$JFILES $f";;
	*.class)	KFILES="$KFILES $f";;
	*.c)		CFILES="$CFILES $f";;
	*.o)		OFILES="$OFILES $f";;
	*.a)		LFILES="$LFILES $f";;
	*.so)		LFILES="$LFILES $f";;
	-l*)		LFILES="$LFILES $f";;
	*)		KFILES="$KFILES $f";;
    esac
done


# exit on any errors beyond here
set -e

# touch timestamp file; arrange deletion on exit
TMP=.tmp.toba.$$
trap 'rm -f $TMP' 0 1 2 15
touch $TMP


# compile Java files to bytecode
INT1=
if [ ${#JFILES} -gt 1 ]; then
    trace export JAVA_HOME=$JDK

trace export CLASSPATH="$CPATH:$JDK/classes:$JDK/lib/classes.zip"

    trace $JAVAC $JFILES
    INT1=
    for f in *.class; do
	if [ ! $f -ot $TMP ]; then
	    INT1="$INT1 $f"
	fi
    done
    KFILES="$KFILES $INT1"
fi

# exit if -S given
$NOTOBA


# translate bytecode to C code
INT2=
if [ ${#KFILES} -gt 1 ]; then
    trace export CLASSPATH="$CPATH:$TOP/classes"
    trace $TOBA $DBGFLAGS $KFILES
    remove $INT1
    INT2=
    for f in *.c; do
	if [ ! $f -ot $TMP ]; then
	    INT2="$INT2 $f"
	fi
    done
    CFILES="$CFILES $INT2"
fi

# exit if -C given
$NOCOMPILE


# compile .c files
INT3=
if [ ${#CFILES} -gt 1 ]; then
    trace $CC $CFLAGS -c $INCLUDES $CFILES
    remove $INT2
    # following "tr" segments prevent sed from getting lines too long
    remove `echo "$INT2 " | tr ' ' '\012' | sed 's/\.c$/.h/g'`
    INT3=`echo "$CFILES " | tr ' ' '\012' | sed 's/\.c$/.o/g'`
    OFILES="$OFILES $INT3"
fi

# exit if -c given
$NOLINK


# link .o files together
# note no space between $RPOPT and $RLIBDIR
LDFLAGS="$LDFLAGS $RPOPT$RLIBDIR"
if [ ${#OFILES} -gt 1 ]; then
    trace $CC $CFLAGS $LDFLAGS -o $XFILE $OFILES $LFILES
    remove $INT3
fi
