/*  runtime.c -- general runtime support routines  */

/* Copyright 1996 Arizona Board of Regents; see COPYRIGHT file for details. */

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "toba.h"
#include "runtime.h"
#include "../gc/gc.h"
#include "sthreads.h"

#include "java_io_PrintStream.h"
#include "java_lang_Error.h"
#include "java_lang_NullPointerException.h"
#include "java_lang_Object.h"
#include "java_lang_String.h"
#include "java_lang_System.h"
#include "java_lang_Thread.h"
#include "java_lang_ThreadDeath.h"



static void report(Object o);
static Object arglist(int argc, char *argv[]);



/*  start(c, main, argc, argv) -- initialize and call main()  */

int start(Class c, Void main(Object o), int argc, char *argv[])
{
    initstructs();			/* initialize structures */
    monitor_package_init();		/* initialize monitor package */
    finalization_init();		/* initialize finalization routines */
    sthread_init();			/* initialize threads package */
    meminit();				/* initialize memory system.
					 * Must happen after threads startup */
    set_main_class(c);			/* initialize API support */

    /* this thread runs main */
    start_thread(main, arglist(argc, argv));

    sthread_exit();                     /* main thread exits */
    terminate(0);			/* do final termination */
    return 1;				/* shouldn't get here */
}

/*
 *  start_thread (func, arg) -- start a thread
 *
 *  At this point the thread should still be marked as unkillable.
 */
void
start_thread(Void (*startfunc)(Object o), Object arg)
{
    struct mythread *thr = mythread();

    /* set up a default exception handler */
    thr->jmpbuf = thr->default_jmpbuf;
    thr->exception_count = 0;
    if (setjmp(thr->jmpbuf)) {		/* exception catcher of last resort */
	sthread_got_exception();
	if (thr->exception == NULL)
	    thr->exception = construct(&cl_java_lang_NullPointerException.C);
	if (instanceof(thr->exception, &cl_java_lang_ThreadDeath.C, 0)) 
	    return;			 /* ignore java.lang.ThreadDeath */
	if (thr->exception_count++ == 0) {  /* try using Java API just once */
	    report(thr->exception);	/* try reporting via API once */
            return;
	} else {
	    fprintf(stderr,"\nuncaught exception\n");
            return;
	}
    }

    /* once we have an exception handler we are killable */
    sthread_dontkill_end(thr);
    startfunc(arg);
}

/*  report(o) -- report uncaught exception o  */

static void report(Object o)
{
    Object s = ((struct in_java_lang_Object*)o)->class->M.toString__4d9OF.f(o);
    fprintf(stderr,"\nuncaught exception: %s\n", cstring(s));
}



/*  arglist(argc, argv) -- construct Java arglist  */

static Object arglist(int argc, char *argv[])
{
    Object o, *a;
    int i;
    
    o = anewarray(&cl_java_lang_String.C, argc - 1);	/* alc String array */
    a = ((struct aarray *)o)->data;
    for (i = 1; i < argc; i++)		/* skip argv[0] */
	a[i - 1] = javastring(argv[i]);
    return o;
}



/*  terminate(status) -- perform any needed exit tasks, and exit  */

void terminate(int status)
{
    memexit();				/* run finalizers */
    exit(status);			/* exit */
}



/*  fatal(mesg) -- immediate fatal error exit  */

void fatal(char *mesg)
{
    fprintf(stderr, "%s\n", mesg);
    exit(1);
}



/*  unimpl(name) -- immediate exit from unimplemented method  */

void unimpl(char *name)
{
    fprintf(stderr, "unimplemented method: %s\n", name);
    exit(1);
}




/*
 *  initclass(myclass) -- initialize class
 *
 *  Elsewhere, myclass->needinit is just checked for zero/nonzero as a boolean.
 *  In here, only, we distinguish between
 *	1  initialization needed
 *	2  initialization in progress
 */
void initclass(Class myclass)
{
    jmp_buf newbuf;
    struct mythread *thr;
    void *oldbuf;
    volatile int monitor_held;

    if (!myclass->needinit)		/* exit quickly if already initialized*/
	return;

    thr = mythread();
    oldbuf = thr->jmpbuf;
    monitor_held = 0;
    if (setjmp(newbuf)) {		/* catch exceptions during init: */
        sthread_got_exception();
        if (monitor_held) {
	    myclass->needinit = 1;	/* mark as uninitialized again */
	    exitclass(myclass, thr, 0, &monitor_held);
        }
	thr->jmpbuf = oldbuf;		/* restore init caller's jump buffer */
	if (instanceof(thr->exception, &cl_java_lang_Error.C, 0))
	    athrow(thr->exception);	/* rethrow error */
	else				/* or mutate other exception */
	    throwExceptionInInitializerError(thr->exception);
    }
    thr->jmpbuf = newbuf;		/* register exception handler */

    /* synchronize vs. other threads */
    enterclass(myclass, thr, 1, &monitor_held);	
    if (!myclass->needinit) {		/* check again under lock */
	exitclass(myclass, thr, 0, &monitor_held);
	thr->jmpbuf = oldbuf;
	return;
    }

    if (myclass->needinit > 1) {
        /* if init is already in progress, must be same thread; just return */
	exitclass(myclass, thr, 0, &monitor_held);
	thr->jmpbuf = oldbuf;
	return;
    }

    myclass->needinit++;		/* indicate initialization in progress*/

    if (myclass->nsupers > 1)		/* if there is a superclass */
	initclass(myclass->supers[1]);	/* initialize it */

    if (myclass->initializer)		/* if there is a <clinit> */
	myclass->initializer();		/* call it */

    myclass->needinit = 0;		/* indicate initialization complete */
    thr->jmpbuf = oldbuf;		/* restore previous exception handler */

    /* release lock */
    exitclass(myclass, thr, 0, &monitor_held);
}



/*
 * cstring(S) -- return a C string corresponding to the java string.
 *
 * characters are truncated to 8 bits; result string is in allocated memory.
 */
char *cstring(struct in_java_lang_String *str)
{
    struct carray *value = (struct carray *)str->value;
    int end, i, j;
    char *cstr;

    cstr = (char *)allocate(str->count + 1);

    end = str->offset + str->count;
    j = 0;
    for(i = str->offset; i < end; i++) {
	cstr[j++] = value->data[i] & 0xff;
    }
    cstr[j] = '\0';

    return cstr;
}



/* javastring(s) -- return a java String class for the ascii C string  */

struct in_java_lang_String *javastring(char *cstr)
{
    struct in_java_lang_String *str;
    struct carray *value;
    int len, i;

    len = strlen(cstr);
    value = anewarray(&cl_char, len);
    for(i = 0; i < len; i++) {
	value->data[i] = cstr[i] & 0xff;
    }

    str = (struct in_java_lang_String *)construct(&cl_java_lang_String.C);
    str->value = value;
    str->offset = 0;
    str->count = len;

    return str;
}



/*  println(o) == System.out.println(o)  (call from dbx to print an object)  */

void println(Object o)
{
    initclass(&cl_java_lang_System.C);
    println_O_n6qWU(cl_java_lang_System.V.out, o);
}

