//  CFile.java -- general procs for writing .c files, excluding method code

//  Copyright 1996 Arizona Board of Regents; see COPYRIGHT file for details.



package toba;

import java.io.*;
import java.util.*;

class CFile {



static private Hashtable declared;	// set of class structs declared
static private BitSet strung;		// set of strings declared

static private int nsupers;		// number of superclasses incl self
static private int ninters;		// number of interfaces implemented
static private int nothers;		// number of "other" classes referenced



//  write(d, c) -- write header info for class c on stream d.

static void write(PrintStream d, ClassInfo c)
{
    declared = new Hashtable();		// clear list of classes

    // include general definitions, present class, ancestors, their interfaces
    d.println();
    d.println("#include \"toba.h\"");
    for (ClassInfo t = c; t != null; t = t.superclass) {
	include(d, t.name);
	for (int i = 0; i < t.interfaces.length; i++)
	    include(d, t.interfaces[i]);
    }

    // always include these two classes (needed by generated code):
    include(d, "java.lang.String");
    include(d, "java.lang.Class");
    
    // include declarations for other referenced classes
    for (int i = 1; i < c.constants.length; i++) {
	Constant k = c.constants[i];
	if (k != null && k.tag == Constant.CLASS) {
	    String s = baseclass((String)k.value);
	    if (s != null)
		include(d, s);
	}
    }

    declared = new Hashtable();		// reset list of classes
    supers(d, c);			// generate superclass list
    inters(d, c);			// generate interfaces list
    others(d, c);			// generate others list

    strung = new BitSet(c.constants.length);
					// initialize list of generated strings

    clstruct(d, c);			// generate class structure

    Constant.gfloats(d, c.constants);	// generate floating constants

    // generate method code
    for (int i = 0; i < c.methods.length; i++) {
	Field m = c.methods[i];
	if ((m.access & ClassInfo.ACC_ABSTRACT) == 0) {
	    // generate body code if not native
	    if ((m.access & ClassInfo.ACC_NATIVE) == 0) {
		try {
		    Method.generate(d, c, m);
		} catch (Error e) {
		    Misc.abort("method " + m.name + ": " + e.getMessage());
		}
	    }
	    // if synchronized, generate wrapper function
	    if ((m.access & ClassInfo.ACC_SYNCHRONIZED) != 0)
		Method.syncwrap(d, c, m);
	}
    }
}



//  include(d, name) -- generate #include for a class, if not already generated

static private void include(PrintStream d, String name)
{
    name = Names.classfile(name);
    if (!declared.containsKey(name)) {
	d.println("#include \"" + name + ".h\"");
	declared.put(name, "");
    }
}



// supers(d, c) -- write list of superclasses.

static private void supers(PrintStream d, ClassInfo c)
{
    nsupers = 0;
    d.println();
    d.println("static const Class supers[] = {");
    for (; c != null; c = c.superclass) {
	d.println("    &cl_" + c.cname + ".C,");
	declared.put(c.cname, "");
	nsupers++;
    }
    d.println("};");
}


// inters(d, c) -- write list of interfaces implemented.

static private void inters(PrintStream d, ClassInfo c)
{
    ninters = 0;
    d.println();
    d.println("static const Class inters[] = {");
    for (; c != null; c = c.superclass) {
	for (int j = 0; j < c.interfaces.length; j++) {
	    String s = Names.hashclass(c.interfaces[j]);
	    d.println("    &cl_" + s + ".C,");
	    declared.put(s, "");
	    ninters++;
	}
    }
    if (ninters == 0)
	d.println("    0");
    d.println("};");
}


// others(d, c) -- write list of other classes referenced.

static private void others(PrintStream d, ClassInfo c)
{
    nothers = 0;
    d.println();
    d.println("static const Class others[] = {");
    for (int i = 1; i < c.constants.length; i++) {
	Constant k = c.constants[i];
	if (k != null && k.tag == Constant.CLASS) {
	    String s = baseclass((String)k.value);
	    if (s != null) {
		s = Names.hashclass(s);
		if (! declared.containsKey(s)) {
		    d.println("    &cl_" + s + ".C,");
		    nothers++;
		}
	    }
	}
    }
    if (nothers == 0)
	d.println("    0");
    d.println("};");
}


//  baseclass(s) -- return substring identifying base class of s.
//
//  If s is an array of primitives, baseclass() returns null.  Otherwise,
//  s is a classname or an array class signature ("[...[Lclassname;"),
//  and baseclass() returns the classname.

static private String baseclass(String s)
{
    int i;

    if (s.charAt(0) != '[')
	return s;
    for (i = 1; s.charAt(i) == '['; i++)
	;
    if (s.charAt(i) == 'L')
	return s.substring(i + 1, s.length() - 1);
    else
	return null;				// baseclass is primitive
}


//  clstruct(d, c) -- write class struct.
//
//  Must be kept in sync with class struct definition in toba.h.

static private void clstruct(PrintStream d, ClassInfo c)
{
    needstr(d, 0, c.name);			// string 0 is class name

    d.println();
    d.println("struct cl_" + c.cname + " cl_" + c.cname + " = {");

    if ((c.access & ClassInfo.ACC_INTERFACE) != 0)
	d.println("    1, 1,");		// needinit, flags (IS_INTERFACE == 1)
    else
	d.println("    1, 0,");

    d.println("    (Object)&s0,");		// class name String constant
    d.println("    &cl_java_lang_Class.C, 0,");	// class Class instance

    d.println("    sizeof(struct in_" + c.cname + "),");
    d.println("    " + c.mtable.length + ",");
    d.println("    " + nsupers + ", supers,");
    d.println("    " + ninters + ", " + c.interfaces.length + ", inters,");
    d.println("    " + nothers + ", others,");
    d.println("    0, 0,");			// arrayclass, elemclass
    d.println("    0,");			// monitor pointer

    methodref(d, c, "<clinit>", false);		// class initializer
    constructor(d, c);				// constructor
    methodref(d, c, "finalize", true);		// finalizer

    methodsigs(d, c);				// methods and signatures

    d.println("};");
}


//  methodref(d, c, name, climb) -- find "()V" method and write its name

static private void methodref
    (PrintStream d, ClassInfo c, String name, boolean climb)
{
    Field m = c.getmethod(name, climb);
    if (m != null)
	d.println("    " + m.cname + ",");
    else
	d.println("    0,");			// not found
}


//  constructor(c, d) -- generate constructor or exception tosser reference

static private void constructor(PrintStream d, ClassInfo c)
{
    if ((c.access & (ClassInfo.ACC_ABSTRACT | ClassInfo.ACC_INTERFACE)) != 0) {
	d.println("    throwInstantiationException,");
	return;
    }
    Field m = c.getmethod("<init>", false);
    if (m == null)
	d.println("    throwNoSuchMethodError,");
    else
	d.println("    " + m.cname + ",");
}



//  methodsigs(d, c) -- generate signatures and references for method table.

static private void methodsigs(PrintStream d, ClassInfo c)
{
    for (int i = 0; i < c.mtable.length; i++) {
	Field m = c.mtable[i];
	if ((m.access & ClassInfo.ACC_ABSTRACT) != 0)
	    d.println("             0, 0,");
	else {
	    String s = String.valueOf(Names.hashinterface(m.name, m.signature));
	    d.println("              ".substring(s.length())
		+ s + ", " + m.cname + ",");
	}
    };
}



//  needstr(d, n, s) -- generate string constant n having value s.
//
//  If the particular constant has already been generated, nothing is done.
//
//  NOTE:  This code is unavoidably tied to the definition of java.lang.String.

static void needstr(PrintStream d, int n, String s)
{
    if (! strung.get(n)) {
	strung.set(n);
	Constant.genstr(d, n, s);
    }
}



} // class CFile
