//  ClassInfo -- the data from a Java class file

//  Copyright 1996 Arizona Board of Regents; see COPYRIGHT file for details.



package toba;

import java.io.*;
import java.util.*;



public class ClassInfo {

// instance variables

    public String name;			// class name
    public String cname;		// name as used in C
    public String fname;		// name used for file
    public String supername;		// superclass name
    public ClassInfo superclass;	// superclass info
    public int major, minor;		// version numbers
    public int access;			// access flags and other flags
    public Constant[] constants;	// constant table
    public String[] interfaces;		// interfaces implemented
    public Field[] fields;		// fields
    public Field[] methods;		// methods
    public Field[] mtable;		// method table (instance methods)
    public Hashtable symtab;		// symbol table (fields & methods)
    public Attribute[] attributes;	// class attributes


// flags from Java class files

public static final int ACC_PUBLIC		= 0x0001;
public static final int ACC_PRIVATE		= 0x0002;
public static final int ACC_PROTECTED		= 0x0004;
public static final int ACC_STATIC		= 0x0008;
public static final int ACC_FINAL		= 0x0010;
public static final int ACC_SYNCHRONIZED	= 0x0020;
public static final int ACC_VOLATILE		= 0x0040;
public static final int ACC_TRANSIENT		= 0x0080;
public static final int ACC_NATIVE		= 0x0100;
public static final int ACC_INTERFACE		= 0x0200;
public static final int ACC_ABSTRACT		= 0x0400;

// additional flags used internally

static final int FLG_HASCV		= 0x10000;	// has class vars


// table of loaded classes

static private Hashtable classtable = new Hashtable();



//  load(name) -- load the data for a class
//
//  Ancestor classes are loaded and linked in as needed.

public static ClassInfo load(String name)
   throws ClassNotFoundException, ClassFormatError, IOException
{
    ClassInfo c;
    DataInputStream d;

    // check for name ending in .class, which indicates a file name
    int len = name.length();
    if (len > 6 && name.substring(len - 6).equals(".class")) {

	// loading by file name
	d = Files.instream(name, Files.classBufferSize);
	if (Debug.selected(Debug.Trace))
	    System.out.println("   loading file " + name);
	name = null;			// don't check class name

    } else {

	// loading by class name; check first to see if already loaded
	c = (ClassInfo)classtable.get(name);
	if (c != null)
	    return c;
	if (Debug.selected(Debug.Trace))
	    System.out.println("   loading class " + name);
	d = Files.findclass(name);

    }

    // load class info from a file
    c = new ClassInfo(d);
    d.close();

    // if loading by class name, check that the right class was found
    if (name != null && !name.equals(c.name))
	throw new ClassNotFoundException(name);

    // remember class for reuse later
    classtable.put(c.name, c);

    // load superclasses
    if (c.supername != null)
	c.superclass = load(c.supername);

    c.scanvars();			// scan variables
    c.scanmeths();			// scan methods

    return c;
}



//  retain() -- keep current class but remove others
//
//  A new classtable is built, retaining only the current class and its
//  superclasses.  Other classes are less likely to be referenced soon,
//  so this saves memory.
//
//  Note:  discarding a class without discarding all its subclasses
//  leaves it occupying memory but effectively unreachable by load().
//  That's worse than useless; watch out for this if modifying the logic.

void retain()
{
    Hashtable h = new Hashtable();
    for (ClassInfo c = this; c != null; c = c.superclass) {
	h.put(c.name, c);
    }
    if (Debug.selected(Debug.Trace)) {
	for (Enumeration e = classtable.keys() ; e.hasMoreElements() ;) {
	    String s = (String)e.nextElement();
	    if (! h.containsKey(s))
		System.out.println("   discarding class " + s);
	}
    }
    classtable = h;
}



//  ClassInfo(d) -- load class data from stream d

ClassInfo(DataInputStream d) throws ClassFormatError, IOException
{
    symtab = new Hashtable();

    // read header
    if (d.readInt() != 0xCafeBabe)
	throw new ClassFormatError("bad magic number");
    minor = d.readUnsignedShort();
    major = d.readUnsignedShort();

    // read constant table
    constants = Constant.load(d);

    // read access flags, class, superclass
    access = d.readUnsignedShort();
    name = (String)constants[d.readUnsignedShort()].value;
    cname = Names.hashclass(name);
    fname = Names.classfile(name);
    int i = d.readUnsignedShort();
    if (i > 0)
	supername = (String)constants[i].value;

    // read interface list
    interfaces = new String[d.readUnsignedShort()];
    for (i = 0; i < interfaces.length; i++) {
	int j = d.readUnsignedShort();
	if (j > 0)
	    interfaces[i] = (String)constants[j].value;
	else
	    interfaces[i] = null;
    }

    // read fields
    fields = new Field[d.readUnsignedShort()];
    for (i = 0; i < fields.length; i++) {
	Field f = new Field(d, this);
	f.next = (Field)symtab.get(f.name);
	symtab.put(f.name, f);
	// for naming purposes, assume no shadowing
	f.cname = Names.hashvar(f.name);
	fields[i] = f;
    }

    // read methods
    methods = new Field[d.readUnsignedShort()];
    for (i = 0; i < methods.length; i++) {
	Field f = new Field(d, this);
	f.next = (Field)symtab.get(f.name);
	symtab.put(f.name, f);
	f.cname = Names.hashmethod(f.name, this.name, f.signature);
	methods[i] = f;
    }

    // read class attributes

    attributes = new Attribute[d.readUnsignedShort()];
    for (i = 0; i < attributes.length; i++)
	attributes[i] = new Attribute(d, this);
}



//  scanvars() -- scan variables.
//
//  Set FLG_HASCV if any class variables are present.

private void scanvars()
{
    for (int i = 0; i < fields.length; i++) {
	if ((fields[i].access & ACC_STATIC) != 0) {
	    access |= FLG_HASCV;		// found a class variable
	    break;
	}
    }
}



//  scanmeths() -- scan methods and build method table.

private void scanmeths()
{
    int mtsize;

    if (superclass == null)
	mtsize = 0;
    else
	mtsize = superclass.mtable.length;

    //  assign method table slots, checking for overrides of ancestors.
nextmethod:
    for (int i = 0; i < methods.length; i++) {
	Field m = methods[i];
	if ((m.access & ACC_STATIC) != 0)
	    continue;		// not virtual (and therefore can't override)

	// this is a virtual method
	String name = m.name;
	int gen = 0;
	for (ClassInfo c = superclass; c != null; c = c.superclass) {
	    gen++;
	    for (Field a = (Field)c.symtab.get(name); a != null; a = a.next) {
		if ((a.access & ACC_PRIVATE) != 0)
		    continue;
		if (m.signature.equals(a.signature)) {
		    // this method overrides an ancestor
		    m.overrides = a.overrides + gen;
		    m.mtslot = a.mtslot;
		    continue nextmethod;
		}
	    }
	}

	// method does not override an ancestor, so it gets a method table slot
	m.mtslot = mtsize++;
    }

    // allocate and initialize method table
    mtable = new Field[mtsize];
    if (superclass != null)
	System.arraycopy(superclass.mtable,0,mtable,0,superclass.mtable.length);
    for (int i = 0; i < methods.length; i++) {
	Field m = methods[i];
	if (m.mtslot >= 0)
	    mtable[m.mtslot] = m;
    }
}



//  getmethod(name, climb) -- find "()V" method by name
//
//  if climb is true, check superclasses if unsuccessful locally.

Field getmethod(String name, boolean climb)
{
    for (ClassInfo c = this; c != null; c = c.superclass) {
	for (Field m = (Field)c.symtab.get(name); m != null; m = m.next) {
	    if (m.signature.equals("()V")) {
		return m;
	    }
	}
	if (!climb)
	    break;
    }
    return null;
}



} // class ClassInfo
