//  Constant -- values from a Java classfile "constant" table

//  Copyright 1996 Arizona Board of Regents; see COPYRIGHT file for details.



package toba;

import java.io.*;
import java.util.*;



public class Constant {		// constant table entry

    public int index;		// entry index
    public int tag;		// constant type
    public Object value;	// constant value


// tag types

public static final int CLASS		=  7;
public static final int FIELD		=  9;
public static final int METHOD		= 10;
public static final int INTERFACE	= 11;
public static final int STRING		=  8;
public static final int INTEGER		=  3;
public static final int FLOAT		=  4;
public static final int LONG		=  5;
public static final int DOUBLE		=  6;
public static final int NAMETYPE	= 12;
public static final int UTF8		=  1;
public static final int UNICODE		=  2;



//  load(d) -- read and crack constant table from stream d.

static Constant[] load (DataInputStream d) throws ClassFormatError, IOException
{
    // allocate constant table; first entry is unused
    Constant[] table = new Constant[d.readUnsignedShort()];
    table[0] = null;

    // read constant table entries
    for (int i = 1; i < table.length; i++) {
	Constant c = new Constant();
	table[i] = c;
	c.index = i;
	c.tag = d.readByte();
	switch (c.tag) {
	    case UTF8:
		c.value = d.readUTF();
		break;
	    case UNICODE:
		int len = d.readUnsignedShort();
		char s[] = new char[len];
		for (int j = 0; j < len; j++)
		    s[j] = (char)d.readUnsignedShort();
		c.value = new String(s);
		break;
	    case INTEGER:
		c.value = new Integer(d.readInt());
		break;
	    case FLOAT:
		c.value = new Float(d.readFloat());
		break;
	    case LONG:
		c.value = new Long(d.readLong());
		table[++i] = null;	// following entry unused
		break;
	    case DOUBLE:
		c.value = new Double(d.readDouble());
		table[++i] = null;	// following entry unused
		break;
	    case CLASS:
	    case STRING:
		c.value = new Integer(d.readUnsignedShort());
		break;
	    case FIELD:
	    case METHOD:
	    case INTERFACE:
	    case NAMETYPE:
		Pair p = new Pair();
		p.i1 = d.readUnsignedShort();
		p.i2 = d.readUnsignedShort();
		c.value = p;
		break;
	    default:
		throw new ClassFormatError("unrecognized constant tag");
	}
    }

    // first, replace string and class pointers with actual values
    for (int i = 1; i < table.length; i++) {
	Constant c = table[i];
	if (c != null) switch (c.tag) {
	    case STRING:
		c.value = table[((Integer)c.value).intValue()].value;
		break;
	    case CLASS:
		String s = (String)table[((Integer)c.value).intValue()].value;
    		c.value = s.replace('/', '.');	// fix path-style classname
		break;
	}
    }

    // then, fill in fields of reference entries
    for (int i = 1; i < table.length; i++) {
	Constant c = table[i];
	if (c != null) switch (c.tag) {
	    case FIELD:
	    case METHOD:
	    case INTERFACE:
		Ref r = new Ref();
		r.pair = (Pair)c.value;
		r.classname = (String)table[r.pair.i1].value;
		Pair p = (Pair)table[r.pair.i2].value;
		r.name = (String)table[p.i1].value;
		r.signature = (String)table[p.i2].value;
		c.value = r;
		break;
	}
    }

    return table;
}



//  dumptable(d, table) -- dump a constant table

static void dumptable(PrintStream d, Constant table[])
{
    d.println();
    d.println("/*  Constant table:");
    for (int i = 1; i < table.length; i++) {
	Constant c = table[i];
	if (c == null)				// if unused entry
	    continue;
	d.print("    c" + i + ".  (" + c.tag + ")  ");
	switch (c.tag) {
	    case FIELD:
	    case METHOD:
	    case INTERFACE:
		Ref r = (Ref)c.value;
		d.println(r.signature + " " + r.classname + "." + r.name);
		break;
	    case NAMETYPE:
		Pair p = (Pair)c.value;
		d.println("c" + p.i1 + ", c" + p.i2);
		break;
	    default:
		d.println(c.value);
		break;
	}
    }
    d.println("*/");
    d.println();
}



//  gfloats(d, table) -- generated needed floating constants for class.

static void gfloats(PrintStream d, Constant table[])
{
    d.println();
    for (int i = 1; i < table.length; i++) {
	Constant c = table[i];
	if (c == null || (c.tag != FLOAT && c.tag != DOUBLE))
	    continue;
	double v = ((Number)c.value).doubleValue();
	if (simplefloat(v))
	    continue;			// will generate in-line value
	if (c.tag == FLOAT)
	    d.println("static union fconst fc" + i + " = { 0x" +
		Integer.toHexString(Float.floatToIntBits((float)v)) + " };");
	else {
	    long b = Double.doubleToLongBits(v);
	    String lh = Integer.toHexString((int)(b >> 32));
	    String rh = Integer.toHexString((int)(b));
	    d.println("static union dconst dc" + i + " = { (ULong)0x" + lh +
		"<<32 | (ULong)(UInt)0x" + rh + " };");
	}
    }
}



//  genstr(d, n, s) -- generate string constant n having value s.

static void genstr(PrintStream d, int n, String s)
{
    int dcllen = s.length();
    if (dcllen == 0)
	dcllen = 1;		// can't declare length-zero array in C
    
    d.println();
    d.print("static const CARRAY(" + dcllen + ") c" + n +
	" = {&acl_char, 0, " + s.length() + ", 0");
    for (int i = 0; i < s.length(); i++) {
	if (i % Misc.GCHARS_PER_LINE == 0 && s.length() > Misc.GCHARS_PER_INIT)
	    d.print(",\n    ");    
	else
	    d.print(",");
	char c = s.charAt(i);
	if (Misc.isQuotable(c))
	    d.print("'" + c + "'");
	else
	    d.print((int)c);
    }
    d.println("};");
    d.println("static const struct in_java_lang_String s" + n +
	" = { &cl_java_lang_String, 0, (Object)&c" +
	n + ", 0, " + s.length() + " };");
}



//  const.repr() -- return C representation of a constant.
//
//  Valid for INT, LONG, FLOAT, DOUBLE, or STRING constants
//  (assuming that necessary static constants were generated earlier).

String repr()
{
    switch (tag) {

	case INTEGER:
	    if (((Integer)value).intValue() == Integer.MIN_VALUE)
		return "0x80000000";
	    else
		return value.toString();

	case LONG:
	    long v = ((Long)value).longValue();
	    if (v > (long)Integer.MIN_VALUE && v < (long)Integer.MAX_VALUE)
		return value.toString();
	    // now we must use "ANSI C" to construct a possibly "long long" val
	    int lh = (int)(v >>> 32);
	    int rh = (int)v;
	    return "((((Long)0x" + Integer.toHexString(lh) +
		"u) << 32) | ((Long)0x" + Integer.toHexString(rh) + "u))";

	case FLOAT:
	case DOUBLE:
	    if (simplefloat(((Number)value).doubleValue()))
		return value.toString();
	    else if (tag == FLOAT)
		return "fc" + index + ".v";
	    else /* tag == DOUBLE */
		return "dc" + index + ".v";

	case STRING:
	    return "(Object)&s" + index;

	default:
	    return value.toString();
    }
}



//  simplefloat(d) -- is d represented accurately by toString()?
//
//  This method errs on the side of caution, but is good enough
//  to accept many exact numbers such as 23.0, 1.5, and 0.125.

static boolean simplefloat(double d)
{
    if (Double.isNaN(d) || Double.isInfinite(d))
	return false;
    if (Double.doubleToLongBits(d) == 0x8000000000000000L)
	return false;			// negative zero

    if (d % 1.0 == 0.0 && d < 1000000.0 && d > -1000000.0)
	return true;
    if (d % 0.03125 == 0.0 && d < 10.0 && d > -10.0)
	return true;

    return false;
}



} // class Constant
