//  Method.java -- method code generation

//  Copyright 1996 Arizona Board of Regents; see COPYRIGHT file for details.



package toba;

import java.io.*;
import java.util.*;



class Method {


//  The ByteCode struct "bc" holds information about the current method.
//  See ByteCode.java.

static ByteCode bc;			// bytecode for current method


//  The generated code needs an encompassing switch() statement if either
//  (1) it catches any exceptions  (bc.handlers.length > 0)
//  (2) it uses JSR/RET

static boolean needswitch;		// need switch() at top?


//  Flags for whether certain things must be generated

static boolean needctemps;		// need tmp class ptrs c0 & c1 declared?
static boolean neednullx;		// need null pointer jump label?


//  Bit vectors record exactly which local variables are used, by JVM type.
//  Only stores are recorded; if it's not stored, it can't be loaded.
//  Don't forget the implicit stores of method parameters, though.

static BitSet[] varused;		// local variables used, by type
static int[] maxvar;			// highest number used, by type


//  JVM types that appear on the stack.

static BitSet stktypes;			// types used on stack


//  Initial assignments of locals from parameters are generated while
//  scanning the method signature, for output later.

static StringBuffer lvinits;		// inits of locals from parameters


//  Return value type in C terms
static String rvtype;			// return value type, if not void



static void generate(PrintStream d, ClassInfo cls, Field m)
{
    int nhandlers;
    int n;

    // initialize
    stktypes = new BitSet();
    maxvar = new int[Misc.CHAR_INDX_SIZE];
    varused = new BitSet[Misc.CHAR_INDX_SIZE];
    for (int i = 0; i < Misc.JVM_TYPES.length(); i++) {
	int c = (int)Misc.JVM_TYPES.charAt(i);
	maxvar[c] = -1;
	varused[c] = new BitSet();
    }

    // if "public static void main(String [])", generate C main entry
    if (m.name.equals("main") && m.signature.equals("([Ljava/lang/String;)V")
    && m.access == (ClassInfo.ACC_PUBLIC | ClassInfo.ACC_STATIC))
	genmain(d, cls, m);

    // find the "Code" attribute
    byte[] cattr = Attribute.find(m.attributes, "Code");
    if (cattr == null)
	Misc.abort("method " + m.name + ": no code");

    // break the attribute into instruction list and auxiliary info
    bc = new ByteCode(cattr);

    nhandlers = bc.handlers.length;	// number of handlers
    needswitch = (nhandlers > 0);	// need switch? (also set by JSR later)
    needctemps = false;			// need class temps? not yet...
    neednullx = false;			// need null pointer catcher? assume no

    // set boundary bits for instructions at ends of exception ranges
    for (int i = 0; i < nhandlers; i++) {
	Handler h = bc.handlers[i];
	bc.instrs[bc.pcmap[h.start]].isBoundary = true;	// mark first instr
	if (h.end < bc.code.length)
	    bc.instrs[bc.pcmap[h.end]].isBoundary = true;  // mark last instr
    }
 
    // number exception range segments, marking each instr by segment number
    if (nhandlers > 0) {
	for (int i = n = 0; i < bc.instrs.length; i++) {
	    Instr ins = bc.instrs[i];
	    if (ins.isBoundary)
		n++;
	    ins.xseg = n;
	}
    }

    // mark directly reachable code, noting jump targets and stack depths
    mark(0, "", 0);		// begin at entry, with empty stack

    // mark exception handling code
    for (int i = 0; i < nhandlers; i++)
	mark(bc.handlers[i].jump, "a", -1);

    // assign labels to jump targets; generate any needed string constants
    d.print("\n\n");
    n = 0;
    for (int i = 0; i < bc.instrs.length; i++) {
	Instr ins = bc.instrs[i];
	if (ins.isTarget)
	    ins.label = n++;
	if (ins.opcode.kind == Opcode.LDC
		&& Main.ctab[ins.val].tag == Constant.STRING)
	    CFile.needstr(d, ins.val, (String)Main.ctab[ins.val].value);
    }

    // dump the code as C comments, if requested
    if (Debug.selected(Debug.Code))
	Instr.dump(d, bc.instrs);

    // generate comment giving hashed and unhashed names
    d.println();
    d.println("/*M " + m.cname + ": " +
	cls.name + "." + m.name + m.signature + " */");

    // generate the function header
    // (n.b. parmdecls() also records other info for later use)
    d.println();
    d.print(Misc.rettype(m) + " ");
    if ((m.access & ClassInfo.ACC_SYNCHRONIZED) != 0)
	d.print("sy_");
    d.println(m.cname + "(" + parmdecls(m) + ")");
    d.println("{");

    // generate the exception table, if any
    if (nhandlers > 0) {
	d.println("static struct handler htable[] = {");
	for (int i = 0; i < nhandlers; i++) {
	    Handler h = bc.handlers[i];
	    if (h.type == 0)
		d.print("    0, ");
	    else
		d.print("    &cl_" +
		    Names.hashclass((String)Main.ctab[h.type].value) + ".C, ");
	    d.println(h.start + ", " + h.end + ", " +
		bc.instrs[bc.pcmap[h.jump]].label + ",");
	}
	d.println("};");
    }

    // declare variables for exception handling and JSR
    if (nhandlers > 0) {
	d.println("struct mythread *tdata;");
	d.println("jmp_buf newbuf;");
	d.println("void *oldbuf;");
	d.println("volatile int pc;");
    }
    if (needswitch)
	d.println("int tgt;");
    if (needswitch && rvtype != null)
	d.println(rvtype + " rv;");

    // declare stack and local variables
    if (needctemps)
	d.println("Class c0, c1;");
    for (int i = 0; i < Misc.JVM_TYPES.length(); i++) {
	char c = Misc.JVM_TYPES.charAt(i);
	dclstack(d, c);
	dclvars(d, c, nhandlers > 0);
    }

    // copy parameters into locals (outputting code generated earlier)
    d.print(lvinits);
    lvinits = null;

    // initialize class, if static function, and if class has initializer
    if ((m.access & ClassInfo.ACC_STATIC) != 0
	    && !m.name.equals("<clinit>")
	    && cls.getmethod("<clinit>",false) != null) {
	d.println();
	d.println("\tif (cl_" + cls.cname + ".C.needinit)");
	d.println("\t\tinitclass(&cl_" + cls.cname + ".C);");
    }

    // initialize exception handling
    if (nhandlers > 0) {
	d.println();
	d.println("\ttdata = mythread();");
	d.println("\toldbuf = tdata->jmpbuf;");
	d.println("\ttgt = 0;");
	d.println("\tif (setjmp(newbuf)) {");
	d.println("\t\tsthread_got_exception();");
	d.println("CATCH:\t\ta1 = tdata->exception;");
	d.println("\t\tif ((tgt = findhandler(htable, " + nhandlers +
	    ", a1, pc)) < 0)");
	d.println("\t\t\tlongjmp(oldbuf, 1);");
	d.println("\t}");
	d.println("\ttdata->jmpbuf = newbuf;");
    } else if (needswitch) {
	d.println();
	d.println("\ttgt = 0;");
    }

    // generate switch top
    if (needswitch) {
	d.println();
	d.println("TOP: switch(tgt) {");
    }

    // generate code for individual instructions
    d.println();
    for (int i = 0; i < bc.instrs.length; i++) {
	Instr ins = bc.instrs[i];
	if (ins.isTarget) {
	    d.print("L" + ins.label + ":");	// generate label
	    if (needswitch)			// and also case label if needed
		d.println("  case " + ins.label + ":");
	}
	if (ins.isBoundary && nhandlers > 0)
	    d.println("\tpc = " + ins.pc + ";");
	if (Debug.selected(Debug.Instrs))
	    d.println("\t\t\t/* " + ins.opcode.name + " */");
	ins.opcode.generate(d, ins);		// generate code for instruction
	if ((ins.opcode.flags & Opcode.NFT) != 0)
		d.println();		// if no fallthrough, add visual break
    }

    // terminate switch
    if (needswitch) {
	d.println("}");
	d.println("RETURN:");
	d.println("\ttdata->jmpbuf = oldbuf;");
	if (rvtype != null)
	    d.println("\treturn rv;");
	else
	    d.println("\treturn;");
    }

    // null pointer exceptions need a jump target
    if (neednullx) {
	d.println("NULLX:");
	d.println("\tthrowNullPointerException(0);");
    }

    d.println("}");
}



//  mark(pc, stack, segnum) -- mark reachable code, maintaining stack state
//
//  If segnum differs from segment number of instruction at pc, the first
//  instruction (at pc) is also marked as a boundary instruction.
//
//  Also marks local variables referenced by STORE-class instructions

private static void mark(int pc, String stack, int segnum)
{
    int i = bc.pcmap[pc];
    Instr ins = bc.instrs[i];
    ins.isTarget = true;
    if (ins.xseg != segnum)
	ins.isBoundary = true;

    while (true) {
	ins = bc.instrs[i++];

	if (ins.isReached) {
	    if (!ins.before.equals(stack))	// sanity check
		throw new VerifyError( "stack mismatch: pc=" + pc +
		   " stack1=" + ins.before + " stack2=" + stack);
	    return;
	}

	ins.isReached = true;
	ins.before = stack;
	if (stack.length() > 0) {
	    char c = stack.charAt(stack.length() - 1);
	    stktypes.set((int)c);
	    if (ins.opcode.kind == Opcode.STORE)
		markvar(c, ins.opcode.var + ins.val);
	}

	stack = ins.after = ins.nextStack();

	if (ins.opcode.kind == Opcode.JSR) {
	    mark(ins.val, stack + "i", ins.xseg); // mark JSR routine with arg
	    mark(ins.pc + ins.length, stack, -1); // mark successor as jump tgt
	    needswitch = true;
	} else if ((ins.opcode.flags & Opcode.PC) != 0)
	    mark(ins.val, stack, ins.xseg);

	if ((ins.opcode.flags & Opcode.SWCH) != 0) {	// if switch instruction
	    mark(ins.more[0], stack, ins.xseg);		// mark default
	    for (int j = 3; j < ins.more.length; j++) {
		mark(ins.more[j], stack, ins.xseg);	// mark other labels
		if (ins.opcode.kind == Opcode.LKPSW)
		    j++;			// skip tag if lookupswitch
	    }
	}

	if ((ins.opcode.flags & Opcode.INST) != 0)
	    needctemps = true;			// need class ptr temps?

	if ((ins.opcode.flags & Opcode.NFT) != 0)
	    break;				// opcode never falls through
	else
	    ins.succ = bc.instrs[i];		// note sequential successor
    }
}



//  parmdecls(m) -- return string of C parameter declarations for Java method m.
//
//  also handles other things while scanning the signature:
//	-- notes local variables needed in "varused" and "maxvar"
//	-- builds list of assignments to those variables in "lvinits"
//	-- notes the return type in "rvtype", if not void

private static String parmdecls(Field m)
{
    String s = m.signature;			// signature being scanned
    int i = 1;					// current position (skip "(")

    StringBuffer b = new StringBuffer();	// parm list being built
    int lv = 0;					// local variable count

    lvinits = new StringBuffer("\n");		// lv initialization code

    if ((m.access & ClassInfo.ACC_STATIC) == 0) {  // instance function?
	b.append("Object p0, ");		// "self" parameter
	parmnote(0, 'a', lv++);
    }

    String t;
loop:
    for (int n = 1; ; n++) {
	char c = s.charAt(i);
	switch (c) {

	    case 'L':				// object
	    default:				// not used, keeps javac happy
		t = "Object";
		i = s.indexOf(';', i);
		parmnote(n, 'a', lv++);
		break;

	    case '[':				// array
		t = "Object";
		while (s.charAt(i) == '[')
		    i++;
		if (s.charAt(i) == 'L')
		    i = s.indexOf(';', i);
		parmnote(n, 'a', lv++);
		break;

	    // primitive types
	    case 'B':  t = "Byte";     parmnote(n, 'i', lv);  lv += 1;  break;
	    case 'C':  t = "Char";     parmnote(n, 'i', lv);  lv += 1;  break;
	    case 'D':  t = "Double";   parmnote(n, 'd', lv);  lv += 2;  break;
	    case 'F':  t = "Float";    parmnote(n, 'f', lv);  lv += 1;  break;
	    case 'I':  t = "Int";      parmnote(n, 'i', lv);  lv += 1;  break;
	    case 'J':  t = "Long";     parmnote(n, 'l', lv);  lv += 2;  break;
	    case 'S':  t = "Short";    parmnote(n, 'i', lv);  lv += 1;  break;
	    case 'Z':  t = "Boolean";  parmnote(n, 'i', lv);  lv += 1;  break;

	    case ')':  break loop;		// end of parameters
	}
	b.append(t).append(" p").append(n).append(", ");
	i++;
    }

    char c = s.charAt(++i);			// return type signature
    if (c == 'V')
	rvtype = null;				// note no return variable
    else
	rvtype = Misc.ctype(c);			// set return variable type

    if (b.length() == 0)
	return "void";				// no parameters
    else {
	b.setLength(b.length() - 2);		// trim final ", " from string
	return b.toString();			// return param declarations
    }
}



//  parmnote(parmnum, jtype, varnum) -- note param and corresponding local var

private static void parmnote(int parmnum, char jtype, int varnum)
{
    markvar(jtype, varnum);
    lvinits.append("\t" + jtype + "v" + varnum + " = p" + parmnum + ";\n");
}



//  markvar(jtype, n) -- mark a local variable as being used.

private static void markvar(char jtype, int n)
{
    int i = (int)jtype;
    if (maxvar[i] < n)
	maxvar[i] = n;
    varused[i].set(n);
}



//  dclstack(d, jtype) -- declare stack variables for a JVM type

static private void dclstack(PrintStream d, char jtype)
{
    if (!stktypes.get((int)jtype))
	return;

    d.print(Misc.ctype(jtype) + " ");
    for (int i = 0; i < bc.max_stack; i++) {
	if (i > 0 && i % Misc.DECLS_PER_LINE == 0)
	    d.print("\n    ");
	d.print(jtype + Integer.toString(i) + ", ");
    }
    d.println(jtype + Integer.toString(bc.max_stack) + ";");
}



//  dclvars(d, jtype, volatile) -- declare local variables marked in bitset

static private void dclvars(PrintStream d, char jtype, boolean vol)
{
    int max = maxvar[(int)jtype];
    if (max < 0)
	return;

    BitSet bs = varused[(int)jtype];
    String s;
    int n = 0;

    for (int i = 0; i <= max; i++) {
	if (bs.get(i)) {
	    s = " " + jtype + "v" + i;
	    if (n % Misc.DECLS_PER_LINE == 0) {
		if (n == 0) {
		    if (vol)
			d.print("volatile ");
    		    d.print(Misc.ctype(jtype) + s);
		} else
	    	    d.print(",\n   " + s);
	    } else
		d.print("," + s);
	    n++;
	}
    }
    d.println(";");
}



//  genmain(d, cls, m) -- generate C main() function for method m

static private void genmain(PrintStream d, ClassInfo cls, Field m)
{
    d.println("\n\n");
    d.println("int main(int argc, char *argv[])\t\t/* main entry */");
    d.println("{");
    d.println("\treturn start(&cl_" + cls.cname + ".C, " +
	m.cname + ", argc, argv);");
    d.println("}");
}



//  syncwrap(d, cls, m) -- generate wrapper function for synchronized method

static void syncwrap(PrintStream d, ClassInfo cls, Field m)
{
    int i, j;
    String p;

    // generate header
    p = parmdecls(m);
    d.println("\n\n");
    d.println("/* wrapper for synchronized method */");
    d.println(Misc.rettype(m) + " " + m.cname + "(" + p + ")");
    d.println("{");

    // declare underlying function, if native
    if ((m.access & ClassInfo.ACC_NATIVE) != 0)
	d.println("\t" + Misc.rettype(m) + " sy_" + m.cname + "();");

    // declare return value, if not void method
    if (rvtype != null)
	d.println("\t" + rvtype + " rv;");

    // declare data pointer, sigsetjmp buffer, and variable
    d.println("\tstruct mythread *tdata;");
    d.println("\tjmp_buf newbuf;");
    d.println("\tvoid *oldbuf;");
    d.println("\tvolatile int monitor_held = 0;");
    d.println("\t");

    // set exception catcher
    d.println("\ttdata = mythread();");
    d.println("\toldbuf = tdata->jmpbuf;");
    d.println("\tif (setjmp(newbuf)) {");
    d.println("\t\tsthread_got_exception();");

    // on exception: exit monitor, if held, and rethrow exception
    d.println("\t\tif (monitor_held)");
    if ((m.access & ClassInfo.ACC_STATIC) != 0) {
        d.println("\t\t\texitclass(&cl_" +
	    cls.cname + ".C, tdata, 0, &monitor_held);");
    } else {
        d.println("\t\t\tmonitorexit(p0, tdata, 0, &monitor_held);");
    }
    d.println("\t\ttdata->jmpbuf = oldbuf;");
    d.println("\t\tathrow(tdata->exception);");
    d.println("\t}");

    // point exception handling to this jump buffer
    d.println("\ttdata->jmpbuf = newbuf;");

    // enter the monitor now that the exception handler is prepared
    if ((m.access & ClassInfo.ACC_STATIC) != 0) {
        // for static functions, the class is locked
        d.println("\tenterclass(&cl_" + 
	    cls.cname + ".C, tdata, 1, &monitor_held);");
    } else {
        // for instance functions, the instance (p0) is locked.
        d.println("\tmonitorenter(p0, tdata, 1, &monitor_held);");
    }

    // call underlying function
    if (rvtype != null)
	d.print("\trv = ");		// assign return value, if any
    else
	d.print("\t");
    d.print("sy_" + m.cname + "(");	// underlying function name
    i = -1;
    while ((i = p.indexOf('p', i + 1)) >= 0) {
	j = p.indexOf(',', i + 1) + 1;	// parameters are pnnn through comma
	if (j <= 0)
	    j = p.length();
	d.print(p.substring(i, j));
    }
    d.println(");");

    // exit monitor
    if ((m.access & ClassInfo.ACC_STATIC) != 0) {
        d.println("\texitclass(&cl_" 
	    + cls.cname + ".C, tdata, 0, &monitor_held);");
    } else {
        d.println("\tmonitorexit(p0, tdata, 0, &monitor_held);");
    }

    // restore old exception handler and return
    d.println("\ttdata->jmpbuf= oldbuf;");
    if (rvtype != null)
	d.println("\treturn rv;");
    d.println("}");
}



} // class Method
